# Copyright (c) 2016-2019 The Regents of the University of Michigan
# This file is part of the General Simulation Data (GSD) project, released under the BSD 2-Clause License.

""" hoomd schema reference implementation

The main package :py:mod:`gsd.hoomd` is a reference implementation of the
GSD schema ``hoomd``. It is a simple, but high performance and memory
efficient, reader and writer for the schema. See :ref:`hoomd-examples`
for full examples.

* :py:func:`create` - Create a hoomd schema GSD file (deprecated).
* :py:func:`open` - Open a hoomd schema GSD file.
* :py:class:`HOOMDTrajectory` - Read and write hoomd schema GSD files.
* :py:class:`Snapshot` - Store the state of a single frame.

    * :py:class:`ConfigurationData` - Store configuration data in a snapshot.
    * :py:class:`ParticleData` - Store particle data in a snapshot.
    * :py:class:`BondData` - Store topology data in a snapshot.
"""

import numpy
from collections import OrderedDict
import logging
import json

try:
    from gsd import fl
except ImportError:
    fl = None;

try:
    import gsd
except ImportError:
    gsd = None;

logger = logging.getLogger('gsd.hoomd')


class ConfigurationData(object):
    """ Store configuration data.

    Users should not need to instantiate this class. Use the ``configuration``
    attribute of a :py:class:`Snapshot`.

    Attributes:
        step (int): Time step of this frame (:chunk:`configuration/step`).
        dimensions (int): Number of dimensions (:chunk:`configuration/dimensions`).
        box (`numpy.ndarray` or `array_like` [float, ndim=1, mode='c']): Box dimensions (:chunk:`configuration/box`)
                                                      - [lx, ly, lz, xy, xz, yz].
    """

    _default_value = OrderedDict();
    _default_value['step'] = numpy.uint64(0);
    _default_value['dimensions'] = numpy.uint8(3);
    _default_value['box'] = numpy.array([1,1,1,0,0,0], dtype=numpy.float32);

    def __init__(self):
        self.step = None;
        self.dimensions = None;
        self.box = None;

    def validate(self):
        """ Validate all attributes.

        First, convert every array attribute to a numpy array of the
        proper type. Then validate that all attributes have the correct
        dimensions.

        Ignore any attributes that are ``None``.

        Warning:
            Array attributes that are not contiguous numpy arrays will
            be replaced with contiguous numpy arrays of the appropriate type.
        """

        logger.debug('Validating ConfigurationData');

        if self.box is not None:
            self.box = numpy.ascontiguousarray(self.box, dtype=numpy.float32);
            self.box = self.box.reshape([6,])

class ParticleData(object):
    """ Store particle data chunks.

    Users should not need to instantiate this class. Use the ``particles``
    attribute of a :py:class:`Snapshot`.

    Instances resulting from file read operations will always store per particle
    quantities in numpy arrays of the defined types. User created snapshots can
    provide input data as python lists, tuples, numpy arrays of different types,
    etc... Such input elements will be converted to the appropriate array type
    by :py:meth:`validate()` which is called when writing a frame.

    Attributes:
        N (int): Number of particles in the snapshot (:chunk:`particles/N`).
        types (list[str]): Names of the particle types (:chunk:`particles/types`).
        position (`numpy.ndarray` or `array_like` [float, ndim=2, mode='c']): Nx3 array defining particle position (:chunk:`particles/position`).
        orientation (`numpy.ndarray` or `array_like` [float, ndim=2, mode='c']): Nx4 array defining particle position (:chunk:`particles/orientation`).
        typeid (`numpy.ndarray` or `array_like` [uint32, ndim=1, mode='c']): N length array defining particle type ids (:chunk:`particles/typeid`).
        mass (`numpy.ndarray` or `array_like` [float, ndim=1, mode='c']): N length array defining particle masses (:chunk:`particles/mass`).
        charge (`numpy.ndarray` or `array_like` [float, ndim=1, mode='c']): N length array defining particle charges (:chunk:`particles/charge`).
        diameter (`numpy.ndarray` or `array_like` [float, ndim=1, mode='c']): N length array defining particle diameters (:chunk:`particles/diameter`).
        body (`numpy.ndarray` or `array_like` [int32, ndim=1, mode='c']): N length array defining particle bodies (:chunk:`particles/body`).
        moment_inertia (`numpy.ndarray` or `array_like` [float, ndim=2, mode='c']): Nx3 array defining particle moments of inertia (:chunk:`particles/moment_inertia`).
        velocity (`numpy.ndarray` or `array_like` [float, ndim=2, mode='c']): Nx3 array defining particle velocities (:chunk:`particles/velocity`).
        angmom (`numpy.ndarray` or `array_like` [float, ndim=2, mode='c']): Nx4 array defining particle angular momenta (:chunk:`particles/angmom`).
        image (`numpy.ndarray` or `array_like` [int32, ndim=2, mode='c']): Nx3 array defining particle images (:chunk:`particles/image`).
        type_shapes (list[dict]): Shape specifications for visualizing particle types (:chunk:`particles/type_shapes`).
    """

    _default_value = OrderedDict();
    _default_value['N'] = numpy.uint32(0);
    _default_value['types'] = ['A'];
    _default_value['typeid'] = numpy.uint32(0);
    _default_value['mass'] = numpy.float32(1.0);
    _default_value['charge'] = numpy.float32(0);
    _default_value['diameter'] = numpy.float32(1.0);
    _default_value['body'] = numpy.int32(-1);
    _default_value['moment_inertia'] = numpy.array([0,0,0], dtype=numpy.float32);
    _default_value['position'] = numpy.array([0,0,0], dtype=numpy.float32);
    _default_value['orientation'] = numpy.array([1,0,0,0], dtype=numpy.float32);
    _default_value['velocity'] = numpy.array([0,0,0], dtype=numpy.float32);
    _default_value['angmom'] = numpy.array([0,0,0,0], dtype=numpy.float32);
    _default_value['image'] = numpy.array([0,0,0], dtype=numpy.int32);
    _default_value['type_shapes'] = [{}];

    def __init__(self):
        self.N = 0;
        self.position = None;
        self.orientation = None;
        self.types = None;
        self.typeid = None;
        self.mass = None;
        self.charge = None;
        self.diameter = None;
        self.body = None;
        self.moment_inertia = None;
        self.velocity = None;
        self.angmom = None;
        self.image = None;
        self.type_shapes = None;

    def validate(self):
        """ Validate all attributes.

        First, convert every per particle attribute to a numpy array of the
        proper type. Then validate that all attributes have the correct
        dimensions.

        Ignore any attributes that are ``None``.

        Warning:
            Per particle attributes that are not contiguous numpy arrays will
            be replaced with contiguous numpy arrays of the appropriate type.
        """

        logger.debug('Validating ParticleData');

        if self.position is not None:
            self.position = numpy.ascontiguousarray(self.position, dtype=numpy.float32);
            self.position = self.position.reshape([self.N, 3])
        if self.orientation is not None:
            self.orientation = numpy.ascontiguousarray(self.orientation, dtype=numpy.float32);
            self.orientation = self.orientation.reshape([self.N, 4])
        if self.typeid is not None:
            self.typeid = numpy.ascontiguousarray(self.typeid, dtype=numpy.uint32);
            self.typeid = self.typeid.reshape([self.N])
        if self.mass is not None:
            self.mass = numpy.ascontiguousarray(self.mass, dtype=numpy.float32);
            self.mass = self.mass.reshape([self.N])
        if self.charge is not None:
            self.charge = numpy.ascontiguousarray(self.charge, dtype=numpy.float32);
            self.charge = self.charge.reshape([self.N])
        if self.diameter is not None:
            self.diameter = numpy.ascontiguousarray(self.diameter, dtype=numpy.float32);
            self.diameter = self.diameter.reshape([self.N])
        if self.body is not None:
            self.body = numpy.ascontiguousarray(self.body, dtype=numpy.int32);
            self.body = self.body.reshape([self.N])
        if self.moment_inertia is not None:
            self.moment_inertia = numpy.ascontiguousarray(self.moment_inertia, dtype=numpy.float32);
            self.moment_inertia = self.moment_inertia.reshape([self.N, 3]);
        if self.velocity is not None:
            self.velocity = numpy.ascontiguousarray(self.velocity, dtype=numpy.float32);
            self.velocity = self.velocity.reshape([self.N, 3]);
        if self.angmom is not None:
            self.angmom = numpy.ascontiguousarray(self.angmom, dtype=numpy.float32);
            self.angmom = self.angmom.reshape([self.N, 4]);
        if self.image is not None:
            self.image = numpy.ascontiguousarray(self.image, dtype=numpy.int32);
            self.image = self.image.reshape([self.N, 3]);

class BondData(object):
    """ Store bond data chunks.

    Users should not need to instantiate this class. Use the ``bonds``,
    ``angles``, ``dihedrals``, or ``impropers`` attribute of a :py:class:`Snapshot`.

    Instances resulting from file read operations will always store per bond
    quantities in numpy arrays of the defined types. User created snapshots can
    provide input data as python lists, tuples, numpy arrays of different types,
    etc... Such input elements will be converted to the appropriate array type
    by :py:meth:`validate()` which is called when writing a frame.

    Note:
        *M* varies depending on the type of bond. The same python class represents all types of bonds.

        ======== ===
        Type     *M*
        ======== ===
        Bond      2
        Angle     3
        Dihedral  4
        Improper  4
        ======== ===

    Attributes:
        N (int): Number of particles in the snapshot (:chunk:`bonds/N`, :chunk:`angles/N`, :chunk:`dihedrals/N`, :chunk:`impropers/N`, :chunk:`pairs/N`).
        types (list[str]): Names of the particle types (:chunk:`bonds/types`, :chunk:`angles/types`, :chunk:`dihedrals/types`, :chunk:`impropers/types`, :chunk:`pairs/types`).
        typeid (`numpy.ndarray` or `array_like` [uint32, ndim=1, mode='c']): - N length array defining bond type ids (:chunk:`bonds/typeid`, :chunk:`angles/typeid`, :chunk:`dihedrals/typeid`, :chunk:`impropers/typeid`, :chunk:`pairs/types`).
        group (`numpy.ndarray` or `array_like` [uint32, ndim=2, mode='c']): - NxM array defining tags in the particle bonds (:chunk:`bonds/group`, :chunk:`angles/group`, :chunk:`dihedrals/group`, :chunk:`impropers/group`, :chunk:`pairs/group`).
    """

    def __init__(self, M):
        self.M = M;
        self.N = 0;
        self.types = None;
        self.typeid = None;
        self.group = None;

        self._default_value = OrderedDict();
        self._default_value['N'] = numpy.uint32(0);
        self._default_value['types'] = [];
        self._default_value['typeid'] = numpy.uint32(0);
        self._default_value['group'] = numpy.array([0]*M, dtype=numpy.int32);

    def validate(self):
        """ Validate all attributes.

        First, convert every per bond attribute to a numpy array of the
        proper type. Then validate that all attributes have the correct
        dimensions.

        Ignore any attributes that are ``None``.

        Warning:
            Per bond attributes that are not contiguous numpy arrays will
            be replaced with contiguous numpy arrays of the appropriate type.
        """

        logger.debug('Validating BondData');

        if self.typeid is not None:
            self.typeid = numpy.ascontiguousarray(self.typeid, dtype=numpy.uint32);
            self.typeid = self.typeid.reshape([self.N])
        if self.group is not None:
            self.group = numpy.ascontiguousarray(self.group, dtype=numpy.int32);
            self.group = self.group.reshape([self.N, self.M]);

class ConstraintData(object):
    """ Store constraint data chunks.

    Users should not need to instantiate this class. Use the ``constraints``,
    attribute of a :py:class:`Snapshot`.

    Instances resulting from file read operations will always store per constraint
    quantities in numpy arrays of the defined types. User created snapshots can
    provide input data as python lists, tuples, numpy arrays of different types,
    etc... Such input elements will be converted to the appropriate array type
    by :py:meth:`validate()` which is called when writing a frame.

    Attributes:
        N (int): Number of particles in the snapshot (:chunk:`constraints/N`).
        value (`numpy.ndarray` or `array_like` [float32, ndim=1, mode='c']): N length array defining constraint lengths (:chunk:`constraints/value`).
        group (`numpy.ndarray` or `array_like` [int32, ndim=2, mode='c']): Nx2 array defining tags in the particle constraints (:chunk:`constraints/group`).
    """

    def __init__(self):
        self.M = 2;
        self.N = 0;
        self.value = None;
        self.group = None;

        self._default_value = OrderedDict();
        self._default_value['N'] = numpy.uint32(0);
        self._default_value['value'] = numpy.float32(0);
        self._default_value['group'] = numpy.array([0]*self.M, dtype=numpy.int32);

    def validate(self):
        """ Validate all attributes.

        First, convert every per constraint attribute to a numpy array of the
        proper type. Then validate that all attributes have the correct
        dimensions.

        Ignore any attributes that are ``None``.

        Warning:
            Per bond attributes that are not contiguous numpy arrays will
            be replaced with contiguous numpy arrays of the appropriate type.
        """

        logger.debug('Validating ConstraintData');

        if self.value is not None:
            self.value = numpy.ascontiguousarray(self.value, dtype=numpy.float32);
            self.value = self.value.reshape([self.N])
        if self.group is not None:
            self.group = numpy.ascontiguousarray(self.group, dtype=numpy.int32);
            self.group = self.group.reshape([self.N, self.M]);

class Snapshot(object):
    """ Top level snapshot container.

    Attributes:
        configuration (:py:class:`ConfigurationData`): Configuration data.
        particles (:py:class:`ParticleData`): Particle data snapshot.
        bonds (:py:class:`BondData`): Bond data snapshot.
        angles (:py:class:`BondData`): Angle data snapshot.
        dihedrals (:py:class:`BondData`): Dihedral data snapshot.
        impropers (:py:class:`BondData`): Improper data snapshot.
        pairs (:py:class:`BondData`): Special pair interactions snapshot
        state (dict): Dictionary containing state data
        log (dict): Dictionary containing logged data (values must be `numpy.ndarray` or `array_like`)

    See the HOOMD schema specification for details on entries in the state dictionary. Entries in this dict are the
    chunk name without the state prefix. For example, :chunk:`state/hpmc/sphere/radius` is stored in the dictionary
    entry ``state['hpmc/sphere/radius']``.
    """

    def __init__(self):
        self.configuration = ConfigurationData();
        self.particles = ParticleData();
        self.bonds = BondData(2);
        self.angles = BondData(3);
        self.dihedrals = BondData(4);
        self.impropers = BondData(4);
        self.constraints = ConstraintData();
        self.pairs = BondData(2);
        self.state = {}
        self.log = {};

        self._valid_state = ['hpmc/integrate/d',
                             'hpmc/integrate/a',
                             'hpmc/sphere/radius',
                             'hpmc/sphere/orientable',
                             'hpmc/ellipsoid/a',
                             'hpmc/ellipsoid/b',
                             'hpmc/ellipsoid/c',
                             'hpmc/convex_polyhedron/N',
                             'hpmc/convex_polyhedron/vertices',
                             'hpmc/convex_spheropolyhedron/N',
                             'hpmc/convex_spheropolyhedron/vertices',
                             'hpmc/convex_spheropolyhedron/sweep_radius',
                             'hpmc/convex_polygon/N',
                             'hpmc/convex_polygon/vertices',
                             'hpmc/convex_spheropolygon/N',
                             'hpmc/convex_spheropolygon/vertices',
                             'hpmc/convex_spheropolygon/sweep_radius',
                             'hpmc/simple_polygon/N',
                             'hpmc/simple_polygon/vertices']

    def validate(self):
        """ Validate all contained snapshot data.
        """

        logger.debug('Validating Snapshot');

        self.configuration.validate();
        self.particles.validate();
        self.bonds.validate();
        self.angles.validate();
        self.dihedrals.validate();
        self.impropers.validate();
        self.constraints.validate();
        self.pairs.validate();

        # validate HPMC state
        if self.particles.types is not None:
            NT = len(self.particles.types)
        else:
            NT = 1;

        if 'hpmc/integrate/d' in self.state:
            self.state['hpmc/integrate/d'] = numpy.ascontiguousarray(self.state['hpmc/integrate/d'], dtype=numpy.float64);
            self.state['hpmc/integrate/d'] = self.state['hpmc/integrate/d'].reshape([1])

        if 'hpmc/integrate/a' in self.state:
            self.state['hpmc/integrate/a'] = numpy.ascontiguousarray(self.state['hpmc/integrate/a'], dtype=numpy.float64);
            self.state['hpmc/integrate/a'] = self.state['hpmc/integrate/a'].reshape([1])

        if 'hpmc/sphere/radius' in self.state:
            self.state['hpmc/sphere/radius'] = numpy.ascontiguousarray(self.state['hpmc/sphere/radius'], dtype=numpy.float32);
            self.state['hpmc/sphere/radius'] = self.state['hpmc/sphere/radius'].reshape([NT])

        if 'hpmc/sphere/orientable' in self.state:
            self.state['hpmc/sphere/orientable'] = numpy.ascontiguousarray(self.state['hpmc/sphere/orientable'], dtype=numpy.uint8);
            self.state['hpmc/sphere/orientable'] = self.state['hpmc/sphere/orientable'].reshape([NT])

        if 'hpmc/ellipsoid/a' in self.state:
            self.state['hpmc/ellipsoid/a'] = numpy.ascontiguousarray(self.state['hpmc/ellipsoid/a'], dtype=numpy.float32);
            self.state['hpmc/ellipsoid/a'] = self.state['hpmc/ellipsoid/a'].reshape([NT])
            self.state['hpmc/ellipsoid/b'] = numpy.ascontiguousarray(self.state['hpmc/ellipsoid/b'], dtype=numpy.float32);
            self.state['hpmc/ellipsoid/b'] = self.state['hpmc/ellipsoid/b'].reshape([NT])
            self.state['hpmc/ellipsoid/c'] = numpy.ascontiguousarray(self.state['hpmc/ellipsoid/c'], dtype=numpy.float32);
            self.state['hpmc/ellipsoid/c'] = self.state['hpmc/ellipsoid/c'].reshape([NT])

        if 'hpmc/convex_polyhedron/N' in self.state:
            self.state['hpmc/convex_polyhedron/N'] = numpy.ascontiguousarray(self.state['hpmc/convex_polyhedron/N'], dtype=numpy.uint32);
            self.state['hpmc/convex_polyhedron/N'] = self.state['hpmc/convex_polyhedron/N'].reshape([NT])
            sumN = numpy.sum(self.state['hpmc/convex_polyhedron/N'])

            self.state['hpmc/convex_polyhedron/vertices'] = numpy.ascontiguousarray(self.state['hpmc/convex_polyhedron/vertices'], dtype=numpy.float32);
            self.state['hpmc/convex_polyhedron/vertices'] = self.state['hpmc/convex_polyhedron/vertices'].reshape([sumN, 3])

        if 'hpmc/convex_spheropolyhedron/N' in self.state:
            self.state['hpmc/convex_spheropolyhedron/N'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolyhedron/N'], dtype=numpy.uint32);
            self.state['hpmc/convex_spheropolyhedron/N'] = self.state['hpmc/convex_spheropolyhedron/N'].reshape([NT])
            sumN = numpy.sum(self.state['hpmc/convex_spheropolyhedron/N'])

            self.state['hpmc/convex_spheropolyhedron/sweep_radius'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolyhedron/sweep_radius'], dtype=numpy.float32);
            self.state['hpmc/convex_spheropolyhedron/sweep_radius'] = self.state['hpmc/convex_spheropolyhedron/sweep_radius'].reshape([NT])

            self.state['hpmc/convex_spheropolyhedron/vertices'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolyhedron/vertices'], dtype=numpy.float32);
            self.state['hpmc/convex_spheropolyhedron/vertices'] = self.state['hpmc/convex_spheropolyhedron/vertices'].reshape([sumN, 3])

        if 'hpmc/convex_polygon/N' in self.state:
            self.state['hpmc/convex_polygon/N'] = numpy.ascontiguousarray(self.state['hpmc/convex_polygon/N'], dtype=numpy.uint32);
            self.state['hpmc/convex_polygon/N'] = self.state['hpmc/convex_polygon/N'].reshape([NT])
            sumN = numpy.sum(self.state['hpmc/convex_polygon/N'])

            self.state['hpmc/convex_polygon/vertices'] = numpy.ascontiguousarray(self.state['hpmc/convex_polygon/vertices'], dtype=numpy.float32);
            self.state['hpmc/convex_polygon/vertices'] = self.state['hpmc/convex_polygon/vertices'].reshape([sumN, 2])

        if 'hpmc/convex_spheropolygon/N' in self.state:
            self.state['hpmc/convex_spheropolygon/N'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolygon/N'], dtype=numpy.uint32);
            self.state['hpmc/convex_spheropolygon/N'] = self.state['hpmc/convex_spheropolygon/N'].reshape([NT])
            sumN = numpy.sum(self.state['hpmc/convex_spheropolygon/N'])

            self.state['hpmc/convex_spheropolygon/sweep_radius'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolygon/sweep_radius'], dtype=numpy.float32);
            self.state['hpmc/convex_spheropolygon/sweep_radius'] = self.state['hpmc/convex_spheropolygon/sweep_radius'].reshape([NT])

            self.state['hpmc/convex_spheropolygon/vertices'] = numpy.ascontiguousarray(self.state['hpmc/convex_spheropolygon/vertices'], dtype=numpy.float32);
            self.state['hpmc/convex_spheropolygon/vertices'] = self.state['hpmc/convex_spheropolygon/vertices'].reshape([sumN, 2])

        if 'hpmc/simple_polygon/N' in self.state:
            self.state['hpmc/simple_polygon/N'] = numpy.ascontiguousarray(self.state['hpmc/simple_polygon/N'], dtype=numpy.uint32);
            self.state['hpmc/simple_polygon/N'] = self.state['hpmc/simple_polygon/N'].reshape([NT])
            sumN = numpy.sum(self.state['hpmc/simple_polygon/N'])

            self.state['hpmc/simple_polygon/vertices'] = numpy.ascontiguousarray(self.state['hpmc/simple_polygon/vertices'], dtype=numpy.float32);
            self.state['hpmc/simple_polygon/vertices'] = self.state['hpmc/simple_polygon/vertices'].reshape([sumN, 2])

        for k in self.state:
            if k not in self._valid_state:
                raise RuntimeError('Not a valid state: ' + k)

class _HOOMDTrajectoryIterable(object):
    """Iterable over a HOOMDTrajectory object."""

    def __init__(self, trajectory, indices):
        self._trajectory = trajectory
        self._indices = indices
        self._indices_iterator = iter(indices)

    def __next__(self):
        return self._trajectory[next(self._indices_iterator)]

    next = __next__     # Python 2.7 compatibility

    def __iter__(self):
        return type(self)(self._trajectory, self._indices)

    def __len__(self):
        return len(self._indices)


class _HOOMDTrajectoryView(object):
    """A view of a HOOMDTrajectory object.

    Enables the slicing and iteration over a subset of a trajectory
    instance.
    """

    def __init__(self, trajectory, indices):
        self._trajectory = trajectory
        self._indices = indices

    def __iter__(self):
        return _HOOMDTrajectoryIterable(self._trajectory, self._indices)

    def __len__(self):
        return len(self._indices)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return type(self)(self._trajectory, self._indices[key])
        else:
            return self._trajectory[self._indices[key]]


class HOOMDTrajectory(object):
    """ Read and write hoomd gsd files.

    Args:
        file (:py:class:`gsd.fl.GSDFile`): File to access.

    Open hoomd GSD files with :py:func:`open`.
    """

    def __init__(self, file):
        self.file = file;
        self._initial_frame = None;

        logger.info('opening HOOMDTrajectory: ' + str(self.file))

        if self.file.schema != 'hoomd':
            raise RuntimeError('GSD file is not a hoomd schema file: ' + str(self.file));
        valid = False;
        if self.file.schema_version < (2,0) and self.file.schema_version >= (1,0):
            valid = True;
        if self.file.schema_version != (0,1):
            valid = True;
        if not valid:
            raise RuntimeError('Incompatible hoomd schema version ' + str(self.file.schema_version) + ' in: ' + str(self.file));

        logger.info('found ' + str(len(self)) + ' frames');

    def __len__(self):
        """ The number of frames in the trajectory. """
        return self.file.nframes;

    def append(self, snapshot):
        """ Append a snapshot to a hoomd gsd file.

        Args:
            snapshot (:py:class:`Snapshot`): Snapshot to append.

        Write the given snapshot to the file at the current frame and
        increase the frame counter. Do not attempt to write any fields
        that are ``None``. For all non-``None`` fields, scan them
        and see if they match the initial frame or the default value.
        If the given data differs, write it out to the frame. If it is
        the same, do not write it out as it can be instantiated either
        from the value at the initial frame or the default value.
        """

        logger.debug('Appending snapshot to hoomd trajectory: ' + str(self.file));

        snapshot.validate();

        # want the initial frame specified as a reference to detect if chunks need to be written
        if self._initial_frame is None and len(self) > 0:
            self.read_frame(0);

        for path in ['configuration', 'particles', 'bonds', 'angles', 'dihedrals', 'impropers', 'constraints', 'pairs']:
            container = getattr(snapshot, path);
            for name in container._default_value:
                if self._should_write(path, name, snapshot):
                    logger.debug('writing data chunk: ' + path + '/' + name);
                    data = getattr(container, name);

                    if name == 'N':
                        data = numpy.array([data], dtype=numpy.uint32);
                    if name == 'step':
                        data = numpy.array([data], dtype=numpy.uint64);
                    if name == 'dimensions':
                        data = numpy.array([data], dtype=numpy.uint8);
                    if name in ('types', 'type_shapes'):
                        if name == 'type_shapes':
                            data = [json.dumps(shape_dict) for shape_dict in data]
                        wid = max(len(w) for w in data)+1;
                        b = numpy.array(data, dtype=numpy.dtype((bytes, wid)));
                        data = b.view(dtype=numpy.int8).reshape(len(b), wid);

                    self.file.write_chunk(path + '/' + name, data)

        # write state data
        for state,data in snapshot.state.items():
            self.file.write_chunk('state/' + state, data)

        # write log data
        for log,data in snapshot.log.items():
            self.file.write_chunk('log/' + log, data)

        self.file.end_frame();

    def truncate(self):
        """ Remove all frames from the file.
        """

        self.file.truncate();
        self._initial_frame = None;

    def _should_write(self, path, name, snapshot):
        """ Test if we should write a given data chunk.

        Args:
            path (str): Path part of the data chunk.
            name (str): Name part of the data chunk.
            snapshot (:py:class:`Snapshot`): Snapshot data is from.
        Returns:
            False if the data matches that in the initial frame. False
            if the data matches all default values. True otherwise.
        """

        container = getattr(snapshot, path);
        data = getattr(container, name);

        if data is None:
            return False;

        if self._initial_frame is not None:
            initial_container = getattr(self._initial_frame, path);
            initial_data = getattr(initial_container, name);
            if numpy.array_equal(initial_data, data):
                logger.debug('skipping data chunk, matches frame 0: ' + path + '/' + name);
                return False;

        if numpy.array_equiv(data, container._default_value[name]):
            logger.debug('skipping data chunk, default value: ' + path + '/' + name);
            return False;

        return True;

    def extend(self, iterable):
        """ Append each item of the iterable to the file.

        Args:
            iterable: An iterable object the provides :py:class:`Snapshot` instances. This could be another
              HOOMDTrajectory, a generator that modifies snapshots, or a simple list of snapshots.
        """

        for item in iterable:
            self.append(item);

    def read_frame(self, idx):
        """ Read the frame at the given index from the file.

        Args:
            idx (int): Frame index to read.
        Returns:
            :py:class:`Snapshot` with the frame data

        Replace any data chunks not present in the given frame with either
        data from frame 0, or initialize from default values if not in
        frame 0. Cache frame 0 data to avoid file read overhead. Return
        any default data as non-writable numpy arrays.
        """

        if idx >= len(self):
            raise IndexError;

        logger.debug('reading frame ' + str(idx) + ' from: ' + str(self.file));

        if self._initial_frame is None and idx != 0:
            self.read_frame(0);

        snap = Snapshot();
        # read configuration first
        if self.file.chunk_exists(frame=idx, name='configuration/step'):
            step_arr = self.file.read_chunk(frame=idx, name='configuration/step');
            snap.configuration.step = step_arr[0];
        else:
            if self._initial_frame is not None:
                snap.configuration.step = self._initial_frame.configuration.step;
            else:
                snap.configuration.step = snap.configuration._default_value['step'];

        if self.file.chunk_exists(frame=idx, name='configuration/dimensions'):
            dimensions_arr = self.file.read_chunk(frame=idx, name='configuration/dimensions');
            snap.configuration.dimensions = dimensions_arr[0];
        else:
            if self._initial_frame is not None:
                snap.configuration.dimensions = self._initial_frame.configuration.dimensions;
            else:
                snap.configuration.dimensions = snap.configuration._default_value['dimensions'];

        if self.file.chunk_exists(frame=idx, name='configuration/box'):
            snap.configuration.box = self.file.read_chunk(frame=idx, name='configuration/box');
        else:
            if self._initial_frame is not None:
                snap.configuration.box = self._initial_frame.configuration.box;
            else:
                snap.configuration.box = snap.configuration._default_value['box'];

        # then read all groups that have N, types, etc...
        for path in ['particles', 'bonds', 'angles', 'dihedrals', 'impropers', 'constraints', 'pairs']:
            container = getattr(snap, path);
            if self._initial_frame is not None:
                initial_frame_container = getattr(self._initial_frame, path);

            container.N = 0;
            if self.file.chunk_exists(frame=idx, name=path+'/N'):
                N_arr = self.file.read_chunk(frame=idx, name=path+'/N');
                container.N = N_arr[0];
            else:
                if self._initial_frame is not None:
                    container.N = initial_frame_container.N;

            # type names
            if 'types' in container._default_value:
                if self.file.chunk_exists(frame=idx, name=path + '/types'):
                    tmp = self.file.read_chunk(frame=idx, name=path + '/types');
                    tmp = tmp.view(dtype=numpy.dtype((bytes, tmp.shape[1])));
                    tmp = tmp.reshape([tmp.shape[0]]);
                    container.types = list(a.decode('UTF-8') for a in tmp)
                else:
                    if self._initial_frame is not None:
                        container.types = initial_frame_container.types;
                    else:
                        container.types = container._default_value['types'];

            # type shapes
            if 'type_shapes' in container._default_value and path == 'particles':
                if self.file.chunk_exists(frame=idx, name=path + '/type_shapes'):
                    tmp = self.file.read_chunk(frame=idx, name=path + '/type_shapes');
                    tmp = tmp.view(dtype=numpy.dtype((bytes, tmp.shape[1])));
                    tmp = tmp.reshape([tmp.shape[0]]);
                    container.type_shapes = list(json.loads(json_string.decode('UTF-8')) for json_string in tmp)
                else:
                    if self._initial_frame is not None:
                        container.type_shapes = initial_frame_container.type_shapes;
                    else:
                        container.type_shapes = container._default_value['type_shapes'];


            for name in container._default_value:
                if name in ('N', 'types', 'type_shapes'):
                    continue;

                # per particle/bond quantities
                if self.file.chunk_exists(frame=idx, name=path + '/' + name):
                    container.__dict__[name] = self.file.read_chunk(frame=idx, name=path + '/' + name);
                else:
                    if self._initial_frame is not None and initial_frame_container.N == container.N:
                        # read default from initial frame
                        container.__dict__[name] = initial_frame_container.__dict__[name];
                    else:
                        # initialize from default value
                        tmp = numpy.array([container._default_value[name]]);
                        s = list(tmp.shape);
                        s[0] = container.N;
                        container.__dict__[name] = numpy.empty(shape=s, dtype=tmp.dtype);
                        container.__dict__[name][:] = tmp;

                    container.__dict__[name].flags.writeable = False;

        # read state data
        for state in snap._valid_state:
            if self.file.chunk_exists(frame=idx, name='state/' + state):
                snap.state[state] = self.file.read_chunk(frame=idx, name='state/' + state);

        # read log data
        logged_data_names = self.file.find_matching_chunk_names('log/')
        for log in logged_data_names:
            if self.file.chunk_exists(frame=idx, name=log):
                snap.log[log[4:]] = self.file.read_chunk(frame=idx, name=log);
            else:
                if self._initial_frame is not None:
                    snap.log[log[4:]] = self._initial_frame.log[log[4:]]

        # store initial frame
        if self._initial_frame is None and idx == 0:
            self._initial_frame = snap;

        return snap;

    def __getitem__(self, key):
        """ Index trajectory frames.

        The index can be a positive integer, negative integer, or slice and is interpreted the same
        as :py:class:`list` indexing.

        Warning:
            As you loop over frames, each frame is read from the file when
            it is reached in the iteration. Multiple passes may lead to
            multiple disk reads if the file does not fit in cache.
        """

        if isinstance(key, slice) :
            return _HOOMDTrajectoryView(self, range(*key.indices(len(self))))
        elif isinstance(key, int) :
            if key < 0:
                key += len(self)
            if key >= len(self) or key < 0:
                raise IndexError();
            return self.read_frame(key);
        else:
            raise TypeError;

    def __iter__(self):
        return _HOOMDTrajectoryIterable(self, range(len(self)))

    def __enter__(self):
        return self;

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close();

def create(name, snapshot=None):
    """ Create a hoomd gsd file from the given snapshot.

    Args:
        name (str): File name.
        snapshot (:py:class:`Snapshot`): Snapshot to write to frame 0. No frame is written if snapshot is ``None``.

    .. deprecated:: 1.2

        As of version 1.2, you can create and open hoomd GSD files in the same call to
        :py:func:`open`. :py:func:`create` is kept for backwards compatibility.

    .. danger::
        The file is overwritten if it already exists.
    """
    if fl is None:
        raise RuntimeError("file layer module is not available");
    if gsd is None:
        raise RuntimeError("gsd module is not available");

    logger.info('creating hoomd gsd file: ' + name);

    gsd.fl.create(name=str(name), application='gsd.hoomd ' + gsd.__version__, schema='hoomd', schema_version=[1,4]);
    with gsd.fl.GSDFile(name, 'wb') as f:
        traj = HOOMDTrajectory(f);
        if snapshot is not None:
            traj.append(snapshot);

def open(name, mode='rb'):
    """ Open a hoomd schema GSD file.

    The return value of :py:func:`open` can be used as a context manager.

    Args:
        name (str): File name to open.
        mode (str): File open mode.

    Returns:
        An :py:class:`HOOMDTrajectory` instance that accesses the file *name* with the given mode.

    Valid values for mode:

    +------------------+---------------------------------------------+
    | mode             | description                                 |
    +==================+=============================================+
    | ``'rb'``         | Open an existing file for reading.          |
    +------------------+---------------------------------------------+
    | ``'rb+'``        | Open an existing file for reading and       |
    |                  | writing. *Inefficient for large files.*     |
    +------------------+---------------------------------------------+
    | ``'wb'``         | Open a file for writing. Creates the file   |
    |                  | if needed, or overwrites an existing file.  |
    +------------------+---------------------------------------------+
    | ``'wb+'``        | Open a file for reading and writing.        |
    |                  | Creates the file if needed, or overwrites   |
    |                  | an existing file. *Inefficient for large    |
    |                  | files.*                                     |
    +------------------+---------------------------------------------+
    | ``'xb'``         | Create a gsd file exclusively and opens it  |
    |                  | for writing.                                |
    |                  | Raise an :py:exc:`FileExistsError`          |
    |                  | exception if it already exists.             |
    +------------------+---------------------------------------------+
    | ``'xb+'``        | Create a gsd file exclusively and opens it  |
    |                  | for reading and writing.                    |
    |                  | Raise an :py:exc:`FileExistsError`          |
    |                  | exception if it already exists.             |
    |                  | *Inefficient for large files.*              |
    +------------------+---------------------------------------------+
    | ``'ab'``         | Open an existing file for writing.          |
    |                  | Does *not* create or overwrite existing     |
    |                  | files.                                      |
    +------------------+---------------------------------------------+

    .. versionadded:: 1.2

    """
    if fl is None:
        raise RuntimeError("file layer module is not available");
    if gsd is None:
        raise RuntimeError("gsd module is not available");

    gsdfileobj = fl.open(name=str(name),
                         mode=mode,
                         application='gsd.hoomd ' + gsd.__version__,
                         schema='hoomd',
                         schema_version=[1,4]);
    return HOOMDTrajectory(gsdfileobj);



# Copyright (c) 2016-2019 The Regents of the University of Michigan
# This file is part of the General Simulation Data (GSD) project, released under the BSD 2-Clause License.

""" GSD file layer API.

Low level access to gsd files. :py:mod:`gsd.fl` allows direct access to create,
read, and write ``gsd`` files. The module is implemented in C and is optimized.
See :ref:`fl-examples` for detailed example code.

* :py:class:`GSDFile` - Class interface to read and write gsd files.
* :py:func:`create` - Create a gsd file (deprecated).
* :py:func:`open` - Open a gsd file.

"""

from libc.stdint cimport uint8_t, int8_t, uint16_t, int16_t, uint32_t, int32_t, uint64_t, int64_t
from libc.errno cimport errno
cimport libgsd
cimport numpy
import os
import logging
import numpy

logger = logging.getLogger('gsd.fl')

########################
### Helper functions ###

cdef __format_errno(fname):
    """ Return a tuple for constructing an IOError """
    return (errno, os.strerror(errno), fname);

# Getter methods for 2D numpy arrays of all supported types
# cython needs strongly typed numpy arrays to get a void *
# to the data, so we implement each by hand here and dispacth
# from the read and write routines
cdef void * __get_ptr_uint8(data):
    cdef numpy.ndarray[uint8_t, ndim=2, mode="c"] data_array_uint8;
    data_array_uint8 = data;
    return <void*>&data_array_uint8[0,0];

cdef void * __get_ptr_uint16(data):
    cdef numpy.ndarray[uint16_t, ndim=2, mode="c"] data_array_uint16;
    data_array_uint16 = data;
    return <void*>&data_array_uint16[0,0];

cdef void * __get_ptr_uint32(data):
    cdef numpy.ndarray[uint32_t, ndim=2, mode="c"] data_array_uint32;
    data_array_uint32 = data;
    return <void*>&data_array_uint32[0,0];

cdef void * __get_ptr_uint64(data):
    cdef numpy.ndarray[uint64_t, ndim=2, mode="c"] data_array_uint64;
    data_array_uint64 = data;
    return <void*>&data_array_uint64[0,0];

cdef void * __get_ptr_int8(data):
    cdef numpy.ndarray[int8_t, ndim=2, mode="c"] data_array_int8;
    data_array_int8 = data;
    return <void*>&data_array_int8[0,0];

cdef void * __get_ptr_int16(data):
    cdef numpy.ndarray[int16_t, ndim=2, mode="c"] data_array_int16;
    data_array_int16 = data;
    return <void*>&data_array_int16[0,0];

cdef void * __get_ptr_int32(data):
    cdef numpy.ndarray[int32_t, ndim=2, mode="c"] data_array_int32;
    data_array_int32 = data;
    return <void*>&data_array_int32[0,0];

cdef void * __get_ptr_int64(data):
    cdef numpy.ndarray[int64_t, ndim=2, mode="c"] data_array_int64;
    data_array_int64 = data;
    return <void*>&data_array_int64[0,0];

cdef void * __get_ptr_float32(data):
    cdef numpy.ndarray[float, ndim=2, mode="c"] data_array_float32;
    data_array_float32 = data;
    return <void*>&data_array_float32[0,0];

cdef void * __get_ptr_float64(data):
    cdef numpy.ndarray[double, ndim=2, mode="c"] data_array_float64;
    data_array_float64 = data;
    return <void*>&data_array_float64[0,0];

def open(name, mode, application, schema, schema_version):
    """ open(name, mode, application, schema, schema_version)

    :py:func:`open` opens a GSD file and returns a :py:class:`GSDFile` instance.
    The return value of :py:func:`open` can be used as a context manager.

    Args:
        name (str): File name to open.
        mode (str): File access mode.
        application (str): Name of the application creating the file.
        schema (str): Name of the data schema.
        schema_version (``list[int]``): Schema version number [major, minor].

    Valid values for mode:

    +------------------+---------------------------------------------+
    | mode             | description                                 |
    +==================+=============================================+
    | ``'rb'``         | Open an existing file for reading.          |
    +------------------+---------------------------------------------+
    | ``'rb+'``        | Open an existing file for reading and       |
    |                  | writing. *Inefficient for large files.*     |
    +------------------+---------------------------------------------+
    | ``'wb'``         | Open a file for writing. Creates the file   |
    |                  | if needed, or overwrites an existing file.  |
    +------------------+---------------------------------------------+
    | ``'wb+'``        | Open a file for reading and writing.        |
    |                  | Creates the file if needed, or overwrites   |
    |                  | an existing file. *Inefficient for large    |
    |                  | files.*                                     |
    +------------------+---------------------------------------------+
    | ``'xb'``         | Create a gsd file exclusively and opens it  |
    |                  | for writing.                                |
    |                  | Raise an :py:exc:`FileExistsError`          |
    |                  | exception if it already exists.             |
    +------------------+---------------------------------------------+
    | ``'xb+'``        | Create a gsd file exclusively and opens it  |
    |                  | for reading and writing.                    |
    |                  | Raise an :py:exc:`FileExistsError`          |
    |                  | exception if it already exists.             |
    |                  | *Inefficient for large files.*              |
    +------------------+---------------------------------------------+
    | ``'ab'``         | Open an existing file for writing.          |
    |                  | Does *not* create or overwrite existing     |
    |                  | files.                                      |
    +------------------+---------------------------------------------+

    The '+' read/write modes are inefficient at handling large files, as they read the entire
    file index into memory. Prefer the appropriate read or write only modes.

    When opening a file for reading (``'r' or 'a'`` modes): ``application`` and ``schema_version`` are ignored.
    :py:func:`open` throws an exception if the file's schema does not match ``schema``.

    When opening a file for writing (``'w' or 'x'`` modes): The given ``application``, ``schema``, and
    ``schema_version`` are saved in the file.

    .. versionadded:: 1.2

    Example:

        .. ipython:: python

            with gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0]) as f:
                f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                f.write_chunk(name='chunk2', data=numpy.array([[5,6],[7,8]], dtype=numpy.float32));
                f.end_frame();
                f.write_chunk(name='chunk1', data=numpy.array([9,10,11,12], dtype=numpy.float32));
                f.write_chunk(name='chunk2', data=numpy.array([[13,14],[15,16]], dtype=numpy.float32));
                f.end_frame();

            f = gsd.fl.GSDFile(name='file.gsd', mode='rb');
            if f.chunk_exists(frame=0, name='chunk1'):
                data = f.read_chunk(frame=0, name='chunk1')
            data
    """

    return GSDFile(str(name), mode, application, schema, schema_version);

cdef class GSDFile:
    """ GSDFile(name, mode, application, schema, schema_version)

    GSD file access interface.

    GSDFile implements an object oriented class interface to the GSD file
    layer. Use :py:func:`open` to open a GSD file and obtain a GSDFile instance.
    :py:class:`GSDFile` can be used as a context manager.

    .. versionchanged:: 1.2

        For new code, use :py:func:`open` instead of constructing GSDFile directly.
        GSDFile.__init__ is backwards compatible with the old open syntax used in GSD versions 1.0.x and 1.1.x.

    Attributes:

        name (str): Name of the open file **(read only)**.
        mode (str): Mode of the open file **(read only)**.
        gsd_version (tuple[int]): GSD file layer version number [major, minor] **(read only)**.
        application (str): Name of the generating application **(read only)**.
        schema (str): Name of the data schema **(read only)**.
        schema_version (tuple[int]): Schema version number [major, minor] **(read only)**.
        nframes (int): Number of frames **(read only)**.
    """

    cdef libgsd.gsd_handle __handle;
    cdef bint __is_open;
    cdef str mode;
    cdef str name;

    def __init__(self, name, mode, application=None, schema=None, schema_version=None):
        cdef libgsd.gsd_open_flag c_flags;
        cdef int exclusive_create = 0;
        cdef int overwrite = 0;

        new_api = (application is not None and schema is not None and schema_version is not None);

        if new_api:
            if mode == 'wb':
                c_flags = libgsd.GSD_OPEN_APPEND;
                overwrite = 1;
            elif mode == 'wb+':
                c_flags = libgsd.GSD_OPEN_READWRITE;
                overwrite = 1;
            elif mode == 'rb':
                c_flags = libgsd.GSD_OPEN_READONLY;
            elif mode == 'rb+':
                c_flags = libgsd.GSD_OPEN_READWRITE;
            elif mode == 'xb':
                c_flags = libgsd.GSD_OPEN_APPEND;
                overwrite = 1;
                exclusive_create = 1;
            elif mode == 'xb+':
                c_flags = libgsd.GSD_OPEN_READWRITE;
                overwrite = 1;
                exclusive_create = 1;
            elif mode == 'ab':
                c_flags = libgsd.GSD_OPEN_APPEND;
            else:
                raise ValueError("mode must be 'wb', 'wb+', 'rb', 'rb+', 'xb', 'xb+', or 'ab'");
        else:
            # backwards compatible old API
            if mode == 'wb':
                c_flags = libgsd.GSD_OPEN_READWRITE;
            elif mode == 'rb':
                c_flags = libgsd.GSD_OPEN_READONLY;
            elif mode == 'ab':
                c_flags = libgsd.GSD_OPEN_APPEND;
            else:
                raise ValueError("mode must be 'rb', 'wb', or 'ab'");

        self.name = name;
        self.mode = mode;

        cdef char * c_name;
        cdef char * c_application;
        cdef char * c_schema;
        cdef int _c_schema_version;
        cdef str schema_truncated;

        if overwrite:
            # create a new file or overwrite an existing one
            logger.info('opening file: ' + name + ' with mode: ' + mode + ', application: ' + application
                         + ', schema: ' + schema + ', and schema_version: ' + str(schema_version));
            name_e = name.encode('utf-8')
            c_name = name_e;

            application_e = application.encode('utf-8')
            c_application = application_e;

            schema_e = schema.encode('utf-8')
            c_schema = schema_e;

            _c_schema_version = libgsd.gsd_make_version(schema_version[0], schema_version[1])

            with nogil:
                retval = libgsd.gsd_create_and_open(&self.__handle,
                                                    c_name,
                                                    c_application,
                                                    c_schema,
                                                    _c_schema_version,
                                                    c_flags,
                                                    exclusive_create);
        else:
            # open an existing file
            logger.info('opening file: ' + name + ' with mode: ' + mode);
            name_e = name.encode('utf-8')
            c_name = name_e;

            with nogil:
                retval = libgsd.gsd_open(&self.__handle, c_name, c_flags);

        if retval == -1:
            raise IOError(*__format_errno(name));
        elif retval == -2:
            raise RuntimeError("Not a GSD file: " + name);
        elif retval == -3:
            raise RuntimeError("Unsupported GSD file version: " + name);
        elif retval == -4:
            raise RuntimeError("Corrupt GSD file: " + name);
        elif retval == -5:
            raise MemoryError("Unable to allocate GSD index: " + name);
        elif retval == -6:
            raise MemoryError("Invalid gsd argument: " + name);
        elif retval != 0:
            raise RuntimeError("Unknown error");

        # validate schema
        if new_api:
            schema_truncated = schema
            if len(schema_truncated) > 64:
                schema_truncated = schema_truncated[0:63]
            if self.schema != schema_truncated:
                raise RuntimeError('file ' + name + ' has incorrect schema: ' + self.schema);

        self.__is_open = True;

    def close(self):
        """ close()

        Close the file.

        Once closed, any other operation on the file object will result in a
        `ValueError`. :py:meth:`close()` may be called more than once.
        The file is automatically closed when garbage collected or when
        the context manager exits.

        Example:
            .. ipython:: python

                f = gsd.fl.open(name='file.gsd', mode='wb+', application="My application", schema="My Schema", schema_version=[1,0])
                f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32))
                f.end_frame();
                data = f.read_chunk(frame=0, name='chunk1')

                f.close()
                # Read fails because the file is closed
                @okexcept
                data = f.read_chunk(frame=0, name='chunk1')

        """
        if self.__is_open:
            logger.info('closing file: ' + self.name);
            with nogil:
                libgsd.gsd_close(&self.__handle);
            self.__is_open = False;

    def truncate(self):
        """ truncate()

        Truncate all data from the file. After truncation, the file has no
        frames and no data chunks. The application, schema, and schema version
        remain the same.

        Example:
            .. ipython:: python

                with gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0]) as f:
                    for i in range(10):
                        f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32))
                        f.end_frame();

                f = gsd.fl.open(name='file.gsd', mode='ab', application="My application", schema="My Schema", schema_version=[1,0])
                f.nframes
                f.schema, f.schema_version, f.application
                f.truncate()
                f.nframes
                f.schema, f.schema_version, f.application
        """

        if not self.__is_open:
            raise ValueError("File is not open");

        logger.info('truncating file: ' + self.name);
        with nogil:
            retval = libgsd.gsd_truncate(&self.__handle);

        if retval == -1:
            raise IOError(*__format_errno(self.name));
        elif retval == -2:
            raise RuntimeError("Not a GSD file: " + self.name);
        elif retval == -3:
            raise RuntimeError("Unsupported GSD file version: " + self.name);
        elif retval == -4:
            raise RuntimeError("Corrupt GSD file: " + self.name);
        elif retval == -5:
            raise MemoryError("Unable to allocate GSD index: " + self.name);
        elif retval != 0:
            raise RuntimeError("Unknown error");


    def end_frame(self):
        """ end_frame()

        Complete writing the current frame. After calling :py:meth:`end_frame()`
        future calls to :py:meth:`write_chunk()` will write to the **next** frame
        in the file.

        .. danger::
            Call :py:meth:`end_frame()` to complete the current frame
            **before** closing the file. If you fail to call
            :py:meth:`end_frame()`, the last frame may not be written
            to disk.

        Example:
            .. ipython:: python

                f = gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0])

                f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                f.end_frame();
                f.write_chunk(name='chunk1', data=numpy.array([9,10,11,12], dtype=numpy.float32));
                f.end_frame();
                f.write_chunk(name='chunk1', data=numpy.array([13,14], dtype=numpy.float32));
                f.end_frame();
                f.nframes

        """

        if not self.__is_open:
            raise ValueError("File is not open");

        logger.debug('end frame: ' + self.name);

        with nogil:
            retval = libgsd.gsd_end_frame(&self.__handle)

        if retval == -1:
            raise IOError(*__format_errno(self.name));
        elif retval == -2:
            raise RuntimeError("GSD file is opened read only: " + self.name);
        elif retval != 0:
            raise RuntimeError("Unknown error");

    def write_chunk(self, name, data):
        """ write_chunk(name, data)

        Write a data chunk to the file. After writing all chunks in the
        current frame, call :py:meth:`end_frame()`.

        Args:
            name (str): Name of the chunk
            data: Data to write into the chunk. Must be a numpy
                  array, or array-like, with 2 or fewer
                  dimensions.

        Warning:
            :py:meth:`write_chunk()` will implicitly converts array-like and
            non-contiguous numpy arrays to contiguous numpy arrays with
            ``numpy.ascontiguousarray(data)``. This may or may not produce
            desired data types in the output file and incurs overhead.

        Example:
            .. ipython:: python

                f = gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0])

                f.write_chunk(name='float1d', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                f.write_chunk(name='float2d', data=numpy.array([[13,14],[15,16],[17,19]], dtype=numpy.float32));
                f.write_chunk(name='double2d', data=numpy.array([[1,4],[5,6],[7,9]], dtype=numpy.float64));
                f.write_chunk(name='int1d', data=numpy.array([70,80,90], dtype=numpy.int64));
                f.end_frame();
                f.nframes
                f.close()
        """

        if not self.__is_open:
            raise ValueError("File is not open");

        data_array = numpy.ascontiguousarray(data);
        if data_array is not data:
            logger.warning('implicit data copy when writing chunk: ' + name);
        data_array = data_array.view();

        cdef uint64_t N;
        cdef uint64_t M;

        if len(data_array.shape) > 2:
            raise ValueError("GSD can only write 1 or 2 dimensional arrays: " + name);

        if len(data_array.shape) == 1:
            data_array = data_array.reshape([data_array.shape[0], 1]);

        N = data_array.shape[0];
        M = data_array.shape[1];

        cdef libgsd.gsd_type gsd_type;
        cdef void *data_ptr;
        if data_array.dtype == numpy.uint8:
            gsd_type = libgsd.GSD_TYPE_UINT8;
            data_ptr = __get_ptr_uint8(data_array)
        elif data_array.dtype == numpy.uint16:
            gsd_type = libgsd.GSD_TYPE_UINT16;
            data_ptr = __get_ptr_uint16(data_array)
        elif data_array.dtype == numpy.uint32:
            gsd_type = libgsd.GSD_TYPE_UINT32;
            data_ptr = __get_ptr_uint32(data_array)
        elif data_array.dtype == numpy.uint64:
            gsd_type = libgsd.GSD_TYPE_UINT64;
            data_ptr = __get_ptr_uint64(data_array)
        elif data_array.dtype == numpy.int8:
            gsd_type = libgsd.GSD_TYPE_INT8;
            data_ptr = __get_ptr_int8(data_array)
        elif data_array.dtype == numpy.int16:
            gsd_type = libgsd.GSD_TYPE_INT16;
            data_ptr = __get_ptr_int16(data_array)
        elif data_array.dtype == numpy.int32:
            gsd_type = libgsd.GSD_TYPE_INT32;
            data_ptr = __get_ptr_int32(data_array)
        elif data_array.dtype == numpy.int64:
            gsd_type = libgsd.GSD_TYPE_INT64;
            data_ptr = __get_ptr_int64(data_array)
        elif data_array.dtype == numpy.float32:
            gsd_type = libgsd.GSD_TYPE_FLOAT;
            data_ptr = __get_ptr_float32(data_array)
        elif data_array.dtype == numpy.float64:
            gsd_type = libgsd.GSD_TYPE_DOUBLE;
            data_ptr = __get_ptr_float64(data_array)
        else:
            raise ValueError("invalid type for chunk: " + name);

        logger.debug('write chunk: ' + self.name + ' - ' + name);

        cdef char * c_name;
        name_e = name.encode('utf-8')
        c_name = name_e;
        with nogil:
            retval = libgsd.gsd_write_chunk(&self.__handle,
                                           c_name,
                                           gsd_type,
                                           N,
                                           M,
                                           0,
                                           data_ptr);

        if retval == -1:
            raise IOError(*__format_errno(name));
        elif retval == -2:
            raise RuntimeError("GSD file is opened read only: " + self.name);
        elif retval == -3:
            raise RuntimeError("Name list full: " + self.name);
        elif retval != 0:
            raise RuntimeError("Unknown error");

    def chunk_exists(self, frame, name):
        """ chunk_exists(frame, name)

        Test if a chunk exists.

        Args:
            frame (int): Index of the frame to check
            name (str): Name of the chunk

        Returns:
            bool: True if the chunk exists in the file. False if it does not.

        Example:
            .. ipython:: python

                with gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0]) as f:
                    f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                    f.write_chunk(name='chunk2', data=numpy.array([[5,6],[7,8]], dtype=numpy.float32));
                    f.end_frame();
                    f.write_chunk(name='chunk1', data=numpy.array([9,10,11,12], dtype=numpy.float32));
                    f.write_chunk(name='chunk2', data=numpy.array([[13,14],[15,16]], dtype=numpy.float32));
                    f.end_frame();

                f = gsd.fl.open(name='file.gsd', mode='rb', application="My application", schema="My Schema", schema_version=[1,0])

                f.chunk_exists(frame=0, name='chunk1')
                f.chunk_exists(frame=0, name='chunk2')
                f.chunk_exists(frame=0, name='chunk3')
                f.chunk_exists(frame=10, name='chunk1')
        """

        cdef const libgsd.gsd_index_entry* index_entry;
        cdef char * c_name;
        name_e = name.encode('utf-8')
        c_name = name_e;
        cdef int64_t c_frame;
        c_frame = frame;

        logger.debug('chunk exists: ' + self.name + ' - ' + name);

        with nogil:
            index_entry = libgsd.gsd_find_chunk(&self.__handle, c_frame, c_name)

        return index_entry != NULL;

    def read_chunk(self, frame, name):
        """ read_chunk(frame, name)

        Read a data chunk from the file and return it as a numpy array.

        Args:
            frame (int): Index of the frame to read
            name (str): Name of the chunk

        Returns:
            ``numpy.ndarray[type, ndim=?, mode='c']``: Data read from file.
            ``type`` is determined by the chunk metadata. If the data is
            NxM in the file and M > 1, return a 2D array. If the data is
            Nx1, return a 1D array.

        .. tip::
            Each call invokes a disk read and allocation of a
            new numpy array for storage. To avoid overhead, don't call
            :py:meth:`read_chunk()` on the same chunk repeatedly. Cache the
            arrays instead.

        Example:
            .. ipython:: python

                with gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0]) as f:
                    f.write_chunk(name='chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                    f.write_chunk(name='chunk2', data=numpy.array([[5,6],[7,8]], dtype=numpy.float32));
                    f.end_frame();
                    f.write_chunk(name='chunk1', data=numpy.array([9,10,11,12], dtype=numpy.float32));
                    f.write_chunk(name='chunk2', data=numpy.array([[13,14],[15,16]], dtype=numpy.float32));
                    f.end_frame();

                f = gsd.fl.open(name='file.gsd', mode='rb', application="My application", schema="My Schema", schema_version=[1,0])
                f.read_chunk(frame=0, name='chunk1')
                f.read_chunk(frame=1, name='chunk1')
                @okexcept
                f.read_chunk(frame=2, name='chunk1')
        """

        if not self.__is_open:
            raise ValueError("File is not open");

        cdef const libgsd.gsd_index_entry* index_entry;
        cdef char * c_name;
        name_e = name.encode('utf-8')
        c_name = name_e;
        cdef int64_t c_frame;
        c_frame = frame;

        with nogil:
            index_entry = libgsd.gsd_find_chunk(&self.__handle, c_frame, c_name)

        if index_entry == NULL:
            raise KeyError("frame " + str(frame) + " / chunk " + name + " not found in: " + self.name);

        cdef libgsd.gsd_type gsd_type;
        gsd_type = <libgsd.gsd_type>index_entry.type;

        cdef void *data_ptr;
        if gsd_type == libgsd.GSD_TYPE_UINT8:
            data_array = numpy.empty(dtype=numpy.uint8, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_UINT16:
            data_array = numpy.empty(dtype=numpy.uint16, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_UINT32:
            data_array = numpy.empty(dtype=numpy.uint32, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_UINT64:
            data_array = numpy.empty(dtype=numpy.uint64, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_INT8:
            data_array = numpy.empty(dtype=numpy.int8, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_INT16:
            data_array = numpy.empty(dtype=numpy.int16, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_INT32:
            data_array = numpy.empty(dtype=numpy.int32, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_INT64:
            data_array = numpy.empty(dtype=numpy.int64, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_FLOAT:
            data_array = numpy.empty(dtype=numpy.float32, shape=[index_entry.N, index_entry.M])
        elif gsd_type == libgsd.GSD_TYPE_DOUBLE:
            data_array = numpy.empty(dtype=numpy.float64, shape=[index_entry.N, index_entry.M])
        else:
            raise ValueError("invalid type for chunk: " + name);

        logger.debug('read chunk: ' + self.name + ' - ' + str(frame) + ' - ' + name);

        # only read chunk if we have data
        if index_entry.N != 0 and index_entry.M != 0:
            if gsd_type == libgsd.GSD_TYPE_UINT8:
                data_ptr = __get_ptr_uint8(data_array)
            elif gsd_type == libgsd.GSD_TYPE_UINT16:
                data_ptr = __get_ptr_uint16(data_array)
            elif gsd_type == libgsd.GSD_TYPE_UINT32:
                data_ptr = __get_ptr_uint32(data_array)
            elif gsd_type == libgsd.GSD_TYPE_UINT64:
                data_ptr = __get_ptr_uint64(data_array)
            elif gsd_type == libgsd.GSD_TYPE_INT8:
                data_ptr = __get_ptr_int8(data_array)
            elif gsd_type == libgsd.GSD_TYPE_INT16:
                data_ptr = __get_ptr_int16(data_array)
            elif gsd_type == libgsd.GSD_TYPE_INT32:
                data_ptr = __get_ptr_int32(data_array)
            elif gsd_type == libgsd.GSD_TYPE_INT64:
                data_ptr = __get_ptr_int64(data_array)
            elif gsd_type == libgsd.GSD_TYPE_FLOAT:
                data_ptr = __get_ptr_float32(data_array)
            elif gsd_type == libgsd.GSD_TYPE_DOUBLE:
                data_ptr = __get_ptr_float64(data_array)
            else:
                raise ValueError("invalid type for chunk: " + name);

            with nogil:
                retval = libgsd.gsd_read_chunk(&self.__handle,
                                               data_ptr,
                                               index_entry);

            if retval == -1:
                raise IOError(*__format_errno(name));
            elif retval == -2:
                raise RuntimeError("Programming error: " + self.name);
            elif retval == -3:
                raise RuntimeError("Corrupt chunk: " + str(frame) + " / " + name + " in file" + self.name);
            elif retval != 0:
                raise RuntimeError("Unknown error");

        if index_entry.M == 1:
            return data_array.reshape([index_entry.N]);
        else:
            return data_array;

    def find_matching_chunk_names(self, match):
        """ find_matching_chunk_names(match)

        Find all the chunk names in the file that start with the string *match*.

        Args:
            match (str): Start of the chunk name to match

        Returns:
            list[str]: Matching chunk names

        Example:
            .. ipython:: python

                with gsd.fl.open(name='file.gsd', mode='wb', application="My application", schema="My Schema", schema_version=[1,0]) as f:
                    f.write_chunk(name='data/chunk1', data=numpy.array([1,2,3,4], dtype=numpy.float32));
                    f.write_chunk(name='data/chunk2', data=numpy.array([[5,6],[7,8]], dtype=numpy.float32));
                    f.write_chunk(name='input/chunk3', data=numpy.array([9, 10], dtype=numpy.float32));
                    f.end_frame();
                    f.write_chunk(name='input/chunk4', data=numpy.array([11, 12, 13, 14], dtype=numpy.float32));
                    f.end_frame();

                f = gsd.fl.open(name='file.gsd', mode='rb', application="My application", schema="My Schema", schema_version=[1,0])

                f.find_matching_chunk_names('')
                f.find_matching_chunk_names('data')
                f.find_matching_chunk_names('input')
                f.find_matching_chunk_names('other')
        """

        if not self.__is_open:
            raise ValueError("File is not open");

        cdef const char * c_found;
        cdef char * c_match;
        match_e = match.encode('utf-8')
        c_match = match_e;

        retval = [];

        with nogil:
            c_found = libgsd.gsd_find_matching_chunk_name(&self.__handle, c_match, NULL);

        while c_found != NULL:
            retval.append(c_found.decode('utf-8'))

            with nogil:
                c_found = libgsd.gsd_find_matching_chunk_name(&self.__handle, c_match, c_found);

        return retval;

    def __enter__(self):
        return self;

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    property name:
        def __get__(self):
            return self.name;

    property mode:
        def __get__(self):
            return self.mode;

    property gsd_version:
        def __get__(self):
            cdef uint32_t v = self.__handle.header.gsd_version;
            return (v >> 16, v & 0xffff);

    property schema_version:
        def __get__(self):
            cdef uint32_t v = self.__handle.header.schema_version;
            return (v >> 16, v & 0xffff);

    property schema:
        def __get__(self):
            return self.__handle.header.schema.decode('utf-8');

    property application:
        def __get__(self):
            return self.__handle.header.application.decode('utf-8');

    property nframes:
        def __get__(self):
            if not self.__is_open:
                raise ValueError("File is not open");

            return libgsd.gsd_get_nframes(&self.__handle);

    def __dealloc__(self):
        if self.__is_open:
            logger.info('closing file: ' + self.name);
            libgsd.gsd_close(&self.__handle);
            self.__is_open = False;

def create(name, application, schema, schema_version):
    """ create(name, application, schema, schema_version)

    Create an empty GSD file on the filesystem.

    .. deprecated:: 1.2

        As of version 1.2, you can create and open GSD files in the same call to
        :py:func:`open`. :py:func:`create` is kept for backwards compatibility.

    Args:
        name (str): File name to open.
        application (str): Name of the application creating the file.
        schema (str): Name of the data schema.
        schema_version (``list[int]``): Schema version number [major, minor].

    Example:

        Create a gsd file:

        .. ipython:: python

            gsd.fl.create(name="file.gsd",
                          application="My application",
                          schema="My Schema",
                          schema_version=[1,0]);

    .. danger::
        The file is overwritten if it already exists.
    """

    _c_schema_version = libgsd.gsd_make_version(schema_version[0], schema_version[1])
    retval = libgsd.gsd_create(str(name).encode('utf-8'), application.encode('utf-8'), schema.encode('utf-8'), _c_schema_version);

    if retval == -1:
        raise IOError(*__format_errno(name));
    elif retval != 0:
        raise RuntimeError("Unknown error");



# Copyright (c) 2016-2019 The Regents of the University of Michigan
# This file is part of the General Simulation Data (GSD) project, released under the BSD 2-Clause License.

""" The GSD main module

The main package :py:mod:`gsd` is the root package. It holds submodules
and does not import them. Users import the modules they need into their python
script::

    import gsd.fl
    f = gsd.fl.GSDFile('filename', 'rb');

Attributes:
    __version__ (str): GSD software version number. This is the version number of the software package as a whole,
                       not the file layer version it reads/writes.
"""

import sys
if sys.version_info < (3, 5) or sys.version_info >= (4, 0):
    raise RuntimeError("Python ~= 3.5 is required")

__version__ = "1.9.1";
